@@ -1,7 +1,7 @@
 #!/usr/bin/env python3
 from __future__ import annotations
 
-__version__ = "1.2.0"
+__version__ = "1.2.1"
 
 import glob
 import sys
@@ -102,12 +102,12 @@ def _mainindex_retrieve(days_threshold=10) -> Optional[MainIndex]:
     """
     picklefile = _MAININDEX_PICKLE_FILE
     if not picklefile.exists():
-        return
+        return None
     import time
     import pickle
     days_since_last_modification = (picklefile.stat().st_mtime - time.time()) / 86400
     if days_since_last_modification > days_threshold:
-        return
+        return None
     f = open(picklefile, "rb")
     try:
         return pickle.load(f)
@@ -207,7 +207,7 @@ class Asset:
     source: str
     """the url where to download the assets from"""
 
-    patterns: List[str] = None
+    patterns: List[str]
     """the paths to extract from the url (if needed). It can be a glob pattern"""
 
     platform: str = 'all'
@@ -410,7 +410,7 @@ class Plugin:
     def __hash__(self):
         return hash((self.name, self.version))
 
-    def binary_filename(self, platform: str = None) -> str:
+    def binary_filename(self, platform: str = None) -> Optional[str]:
         """
         The filename of the binary (a .so, .dll or .dylib file)
 
@@ -440,7 +440,7 @@ class Plugin:
         path = self.resolve_doc_folder() / markdownfile
         return path if path.exists() else None
 
-    def resolve_doc_folder(self) -> Optional[Path]:
+    def resolve_doc_folder(self) -> Path:
         """
         Resolve the doc folder for this plugin
 
@@ -481,6 +481,10 @@ class InstalledPluginInfo:
     installed_manifest_path: Optional[Path] = None
     installed_in_system_folder: bool = False
 
+    @property
+    def versiontuple(self) -> Tuple[int, int, int]:
+        return _version_tuple(self.versionstr) if self.versionstr else (0, 0, 0)
+
 
 UNKNOWN_VERSION = "Unknown"
 
@@ -511,7 +515,7 @@ def _main_repository_path() -> Path:
     return RISSET_ROOT / "risset-data"
 
 
-def user_plugins_path() -> Optional[Path]:
+def user_plugins_path() -> Path:
     """
     Return the install path for user plugins
 
@@ -636,6 +640,8 @@ def _csound_opcodes() -> List[str]:
     Returns a list of installed opcodes
     """
     csound_bin = _get_binary("csound")
+    if not csound_bin:
+        raise RuntimeError("Did not find csound binary")
     proc = subprocess.run([csound_bin, "-z1"], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
     txt = proc.stdout.decode('ascii')
     opcodes = []
@@ -760,8 +766,8 @@ def _git_repo_needs_update(repopath: Path) -> bool:
     os.chdir(str(repopath))
     git = _get_git_binary()
     subprocess.call([git, "fetch"])
-    headhash = subprocess.check_output([git, "rev-parse", "HEAD"])
-    upstreamhash =subprocess.check_output([git, "rev-parse", "master@{upstream}"])
+    headhash = subprocess.check_output([git, "rev-parse", "HEAD"]).decode('utf-8')
+    upstreamhash =subprocess.check_output([git, "rev-parse", "master@{upstream}"]).decode('utf-8')
     _debug(f"Checking hashes, head: {headhash}, upstream: {upstreamhash}")
     os.chdir(cwd)
     return headhash != upstreamhash
@@ -790,7 +796,7 @@ def _git_update(repopath: Path, depth=0, check_if_needed=False) -> None:
     os.chdir(cwd)
 
 
-def _version_tuplet(versionstr: str) -> Tuple[int, int, int]:
+def _version_tuple(versionstr: str) -> Tuple[int, int, int]:
     """ Convert a version string to its integer parts """
     if not versionstr:
         raise ValueError("versionstr is empty")
@@ -887,7 +893,7 @@ def _parse_pluginkey(pluginkey: str) -> Tuple[str, str]:
 
 def _normalize_version(version: str, default="0.0.0") -> str:
     try:
-        versiontup = _version_tuplet(version)
+        versiontup = _version_tuple(version)
     except ValueError as e:
         _debug(f"Error while parsing version {version}: %s", str(e))
         return default
@@ -911,7 +917,7 @@ def _parse_asset(assetdef: dict, defaultsource: str) -> Asset:
     extractpath = assetdef.get('extractpath') or assetdef.get('path')
     if not source and not extractpath:
         raise ParseError(f"Asset definition should have an URL or an extractpath key")
-    paths = extractpath.split(";") if extractpath else None
+    paths = extractpath.split(";") if extractpath else []
     return Asset(source=source, patterns=paths, platform=assetdef.get('platform', 'all'), name=assetdef.get('name', ''))
 
 
@@ -1085,7 +1091,7 @@ def _make_install_manifest(plugin: Plugin, assetfiles: List[str] = None) -> dict
             under the plugins prefix)
     """
     platform = _get_platform()
-    out = {}
+    out: Dict[str, Any] = {}
     out['name'] = plugin.name
     out['author'] = plugin.author
     out['email'] = plugin.email
@@ -1160,7 +1166,6 @@ def system_plugins_path() -> Optional[Path]:
     if not out:
         _errormsg(f"System plugins path not found! Searched paths: {possible_paths}")
         return None
-    assert isinstance(out, Path)
     assert out.exists() and out.is_dir() and out.is_absolute()
     register.system_plugins_path = out
     return out
@@ -1170,22 +1175,27 @@ def user_installed_dlls() -> List[Path]:
     """
     Return a list of plugins installed at the user plugin path.
     """
+    if cached:=register.cache.get('user_installed_dlls'):
+        return cached
     path = user_plugins_path()
     if not path or not path.exists():
         return []
-    ext = _plugin_extension()
-    return list(path.glob("*"+ext))
+    register.cache['user_installed_dlls'] = out = list(path.glob("*"+_plugin_extension()))
+    return out
 
 
 def system_installed_dlls() -> List[Path]:
     """
     LIst of plugins installed at the system's path
     """
+    if cached:=register.cache.get('system_installed_dlls'):
+        return cached
     path = system_plugins_path()
     if not path or not path.exists():
         return []
     ext = _plugin_extension()
-    return list(path.glob("*" + ext))
+    register.cache['system_installed_dlls'] = out = list(path.glob("*" + ext))
+    return out
 
 
 class MainIndex:
@@ -1307,7 +1317,7 @@ class MainIndex:
         """
         return _generate_documentation(self, dest=dest, buildhtml=buildhtml, onlyinstalled=onlyinstalled)
 
-    def _parse_plugin(self, pluginname: str) -> Optional[Plugin]:
+    def _parse_plugin(self, pluginname: str) -> Plugin:
         pluginsource = self.pluginsources.get(pluginname)
         if pluginsource is None:
             raise KeyError(f"Plugin {pluginname} not known. Known plugins: {self.pluginsources.keys()}")
@@ -1322,6 +1332,19 @@ class MainIndex:
             raise err
         return pluginsource.read_definition()
 
+    def installed_dlls(self) -> Dict[str, Tuple[Path, bool]]:
+        if cached:=self._cache.get('installed_dlls'):
+            return cached
+        user_dlls = user_installed_dlls()
+        system_dlls = system_installed_dlls()
+        db = {}
+        for dll in user_dlls:
+            db[dll.name] = (dll, True)
+        for dll in system_dlls:
+            db[dll.name] = (dll, False)
+        self._cache['installed_dlls'] = db
+        return db
+
     def installed_path_for_dll(self, binary: str) -> Tuple[Optional[Path], bool]:
         """
         Get the installed path for a given plugin binary
@@ -1335,15 +1358,12 @@ class MainIndex:
         Returns:
             A tuple (path to the actual file or None if not found, True if this is inside the user plugins path)
         """
-        user_dlls = user_installed_dlls()
-        for user_dll in user_dlls:
-            if user_dll.name == binary:
-                return user_dll, True
-        system_dlls = system_installed_dlls()
-        for system_dll in system_dlls:
-            if system_dll.name == binary:
-                return system_dll, False
-        return None, False
+        dlldb = self.installed_dlls()
+        if binary in dlldb:
+            path, userinstalled = dlldb[binary]
+            return path, userinstalled
+        else:
+            return None, False
 
     def installed_manifests_path(self) -> Path:
         """
@@ -1387,7 +1407,10 @@ class MainIndex:
             check: if True, we check if the opcodes declared in the plugin definition
                 are actually available
         """
-        dll, user_installed = self.installed_path_for_dll(plugin.binary_filename())
+        binfile = plugin.binary_filename()
+        if not binfile:
+            return False
+        dll, user_installed = self.installed_path_for_dll(binfile)
         return dll is not None and (not check or self._check_plugin_installed(plugin))
 
     def find_manpage(self,
@@ -1419,13 +1442,18 @@ class MainIndex:
                 if opcode in plugin.opcodes:
                     return plugin.manpage(opcode)
             _errormsg(f"Opcode {opcode} not found")
+            return None
 
     def installed_plugin_info(self, plugin: Plugin) -> Optional[InstalledPluginInfo]:
         """
         Returns an InstalledPluginInfo if found, None otherwise
         """
         _debug(f"Checking if plugin {plugin.name} is installed")
-        dll, user_installed = self.installed_path_for_dll(plugin.binary_filename())
+        binfile = plugin.binary_filename()
+        if not binfile:
+            _debug(f"No binary for this platform (plugin {plugin.name} supports platforms {plugin.binaries.keys()}")
+            return None
+        dll, user_installed = self.installed_path_for_dll(binfile)
         if not dll:
             # plugin is not installed
             _debug(f"plugin {plugin.name} is not installed yet")
@@ -1510,22 +1538,6 @@ class MainIndex:
         else:
             raise SchemaError(f"Suffix {path.suffix} not supported in url: {bindef.url}")
 
-    def plugin_installed_version(self, plugin: Plugin) -> Optional[str]:
-        """
-        Check if the plugin is installed, return its version as string
-
-        A valid version has the form <major>.<minor>[.<patch>] (patch is optional)
-
-        Returns None if the dll is not installed, UNKNOWN_VERSION if the
-        dll is installed but there is no corresponding installation
-        manifest (it was not installed via risset)
-        """
-        info = self.installed_plugin_info(plugin)
-        if not info:
-            _debug(f"Plugin {plugin.name} is not installed")
-            return None
-        return info.versionstr
-
     def opcodes_by_name(self) -> Dict[str, Opcode]:
         """
         Returns a dict mapping opcodename to an Opcode definition
@@ -1540,7 +1552,7 @@ class MainIndex:
 
     def parse_manpage(self, opcode: str) -> Optional[ManPage]:
         manpage = self.find_manpage(opcode, markdown=True)
-        return _manpage_parse(manpage, opcode)
+        return _manpage_parse(manpage, opcode) if manpage else None
 
     def defined_opcodes(self) -> List[Opcode]:
         """
@@ -1551,18 +1563,23 @@ class MainIndex:
             return cached
         opcodes = []
         for plugin in self.plugins.values():
-            pluginstalled = self.is_plugin_installed(plugin)
+            pluginstalled = self.is_plugin_installed(plugin, check=False)
             for opcodename in plugin.opcodes:
                 manpage = self.parse_manpage(opcodename)
-                opcodes.append(Opcode(name=opcodename, plugin=plugin.name, abstract=manpage.abstract,
-                                      syntaxes=manpage.syntaxes, installed=pluginstalled))
+                if not manpage:
+                    _errormsg(f"No manpage for opcode {opcodename}!")
+                opcodes.append(Opcode(name=opcodename, plugin=plugin.name, installed=pluginstalled,
+                                      abstract=manpage.abstract if manpage else '?',
+                                      syntaxes=manpage.syntaxes if manpage else []))
         opcodes.sort(key=lambda opc: opc.name.lower())
         self._cache['opcodes'] = opcodes
         return opcodes
 
-    def serialize(self, outfile: str = _MAININDEX_PICKLE_FILE):
+    def serialize(self, outfile: Union[str, Path] = _MAININDEX_PICKLE_FILE) -> None:
         import pickle
-        db = self.opcodes_by_name()
+        # Populate cache
+        _ = self.opcodes_by_name()
+        _ = self.installed_dlls()
         pickle.dump(self, open(outfile, 'wb'))
 
     def install_plugin(self, plugin: Plugin, check=False) -> Optional[ErrorMsg]:
@@ -1602,7 +1619,7 @@ class MainIndex:
         if not self.is_plugin_installed(plugin, check=check):
             if not check:
                 return ErrorMsg(f"Tried to install plugin {plugin.name}, but the binary"
-                                f"is not present")
+                                f" is not present")
             else:
                 return ErrorMsg(f"Tried to install plugin {plugin.name}, but opcode "
                                 f"{plugin.opcodes[0]}, which is provided by this plugin, "
@@ -1646,7 +1663,7 @@ class MainIndex:
             plugininstalled = info is not None
             if installed and not plugininstalled:
                 continue
-            plugdict = {}
+            plugdict: Dict[str, Any] = {}
             plugdict['version'] = plugin.version
             if info:
                 plugdict['installed'] = True
@@ -1791,7 +1808,7 @@ class MainIndex:
                         os.remove(assetfullpath)
                 remainingassets = list(assetsfolder.glob("*"))
                 if remainingassets:
-                    _info(f"There are remaining assets in the folder {assetsfolder}: {remainingassets}")
+                    _info(f"There are remaining assets in the folder {assetsfolder}: {', '.join(p.as_posix() for p in remainingassets)}")
                     _info("... They will be removed")
                 _rm_dir(assetsfolder)
             os.remove(manifestpath.as_posix())
@@ -1968,7 +1985,7 @@ def _compile_docs(index: MainIndex, dest: Path, makeindex=True,
     shutil.copy(syntaxhighlightingcss, css_folder)
 
     for plugin in index.plugins.values():
-        if onlyinstalled and not index.is_plugin_installed(plugin):
+        if onlyinstalled and not index.is_plugin_installed(plugin, check=False):
             continue
         doc_folder = plugin.resolve_doc_folder()
         if doc_folder is None:
@@ -2051,7 +2068,7 @@ def _manpage_parse(manpage: Path, opcode: str) -> Optional[ManPage]:
 
 
 def _docs_generate_index(index: MainIndex,
-                         outfile: Path = None) -> None:
+                         outfile: Path) -> None:
     """
     Generate an index for the documentation
 
@@ -2059,7 +2076,7 @@ def _docs_generate_index(index: MainIndex,
         index: the main index
         outfile: the path to write the index to (normally an index.md file)
     """
-    lines = []
+    lines: List[str] = []
     _ = lines.append
     _("# Plugins\n")
     plugins = sorted(index.plugins.values(), key=lambda plugin: plugin.name)
@@ -2115,17 +2132,19 @@ def cmd_rm(index: MainIndex, args) -> bool:
     """
     Remove a plugin
     """
-    errors = []
-    platform = _get_platform()
-    noerrors = True
+    errors = 0
     for pluginname in args.plugin:
         plugdef = index.plugins.get(pluginname)
+        if not plugdef:
+            _errormsg(f"Plugin {pluginname} not defined. Known plugins: {', '.join(index.plugins.keys())}")
+            errors +=1
+            continue
         try:
             index.uninstall_plugin(plugdef)
         except Exception as e:
             _errormsg(str(e))
-            noerrors = False
-    return noerrors
+            errors += 1
+    return errors == 0
 
 
 def cmd_install(index: MainIndex, args) -> bool:
@@ -2156,25 +2175,25 @@ def cmd_install(index: MainIndex, args) -> bool:
     allplugins = list(set(allplugins))  # remove duplicates
     errors_found = False
     for plugin in allplugins:
-        current_version = index.plugin_installed_version(plugin)
-        if current_version == UNKNOWN_VERSION:
+        plugininfo = index.installed_plugin_info(plugin)
+        if not plugininfo:
+            # Not installed
+            _debug(f"Plugin {plugin} not installed, installing")
+        elif not plugininfo.versionstr or plugininfo.versionstr == UNKNOWN_VERSION:
             # plugin is installed but without a corresponding install manifest.
             if not args.force:
                 _errormsg(f"Plugin {plugin} is already installed. Use --force to force reinstall")
                 errors_found = True
                 continue
-        elif current_version is None:
-            # plugin is not installed
-            _debug(f"Plugin {plugin} not installed, installing")
         else:
-            if _version_tuplet(plugin.version) <= _version_tuplet(current_version):
+            if _version_tuple(plugin.version) <= plugininfo.versiontuple:
                 _debug(f"Plugin {plugin.name}, version: {plugin.version}")
-                _debug(f"    Installed version: {current_version}")
+                _debug(f"    Installed version: {plugininfo.versionstr}")
                 _info(f"Installed version of plugin {plugin.name} is up-to-date")
                 errors_found = True
                 continue
             _info(f"Updating plugin {plugin.name}: "
-                 f"{current_version} -> {plugin.version}")
+                 f"{plugininfo.versionstr} -> {plugin.version}")
         error = index.install_plugin(plugin)
         if error:
             _errormsg(error)
@@ -2210,7 +2229,7 @@ def cmd_man(idx: MainIndex, args) -> bool:
                       the markdown is output to the terminal
         opcode      - opcode(s) to get manpage of. Can be a wildcard
     """
-    opcodes = []
+    opcodes: List[Opcode] = []
     if args.html and not args.markdown:
         fmt = "html"
     else:
@@ -2318,8 +2337,8 @@ def cmd_info(idx: MainIndex, args) -> bool:
         'datarepo': RISSET_DATAREPO_LOCALPATH.as_posix(),
         'opcodesxml': (RISSET_ROOT / "opcodes.xml").as_posix(),
         'days-since-update': lastupdate,
-        'installed-plugins': [plugin for plugin in idx.plugins.values()
-                              if idx.is_plugin_installed()]
+        'installed-plugins': [plugin.name for plugin in idx.plugins.values()
+                              if idx.is_plugin_installed(plugin, check=False)]
     }
     jsonstr = json.dumps(d, indent=True)
     if args.outfile:
